﻿using System;
using System.Diagnostics;

namespace DF
{
	public class Noeud
	{
		public static int compteur;

		public Position p;
		public Noeud pere;
		public Noeud[] fils;
		public int cross, win;
		public int indiceMeilleurFils;

		public Noeud(Noeud pere, Position p)
		{
			this.pere = pere;
			this.p = p;
			fils = new Noeud[p.NbCoups];
			compteur++;
		}

		public void CalculMeilleurFils(float a, float b)
		{
			//if (fils.Length == 0) return;
			indiceMeilleurFils = JMCTSinf.gen.Next(fils.Length);
			float s;
			if (p.j1aletrait)
			{
				float sM = fils[indiceMeilleurFils] == null ? 1F : (fils[indiceMeilleurFils].win + a) / (fils[indiceMeilleurFils].cross + b);
				for (int i = 0; i < fils.Length; i++)
				{
					s = fils[i] == null ? 1F : (fils[i].win + a) / (fils[i].cross + b);
					if (s > sM) { sM = s; indiceMeilleurFils = i; }
				}
			}
			else {
				float sM = fils[indiceMeilleurFils] == null ? 1F : ((fils[indiceMeilleurFils].cross - fils[indiceMeilleurFils].win) + a) / (fils[indiceMeilleurFils].cross + b);
				for (int i = 0; i < fils.Length; i++)
				{
					s = fils[i] == null ? 1F : ((fils[i].cross - fils[i].win + a)) / (fils[i].cross + b);
					if (s > sM) { sM = s; indiceMeilleurFils = i; }
				}
			}
		}

		public Noeud MeilleurFils()
		{
			return InstancierFils(indiceMeilleurFils);
		}


		public Noeud InstancierFils(int i)
		{
			if (fils[i] != null) { return fils[i]; }
			fils[i] = new Noeud(this, p.PositionFille(i));
			return fils[i];
		}

		public override string ToString()
		{
			string s = "Win/Cross=" + win + "/" + cross + "\nFils : ";
			for (int k = 0; k < fils.Length; k++)
			{
				if (fils[k] != null)
				{
					s += (fils[k].win + "/" + fils[k].cross + " ");
				}
				else s += (0 + "/" + 0 + " ");
			}
			return s;
		}
	}

	public class JMCTSinf : Joueur
	{
		public static Random gen = new Random();
		static Stopwatch sw = new Stopwatch();

		float a,b;
		int n;
		int temps;

		int nbiter, nbcoups;

		public override float Nbitmoy
		{
			get
			{
				return (1F * nbiter) / nbcoups;
			}
		}

		Noeud racine;

		public JMCTSinf(float a, float b, int n, int temps=int.MaxValue)
		{
			this.a = a; this.b = b;  this.n = n; this.temps = temps;
		}

		public override string ToString()
		{
			return string.Format("JMCTSinf[a={0}  b={1}  n={2}  temps={3}]", a, b, n, temps);
		}

		public override void NouvellePartie()
		{
			racine=null;
			nbiter = 0;
			nbcoups = 0;
		}

		public override int Jouer(Position p, bool affichage)
		{
			Noeud.compteur = 0;
			sw.Restart();

			if (racine == null) { racine = new Noeud(null, p); }
			else {
				int k = 0;
				for (k = 0; k < racine.fils.Length; k++)
				{
					if (p.Equals(racine.p.PositionFille(k))) break;
				}
				racine = racine.InstancierFils(k);
				racine.pere = null;
			}

			int i = 0;
			for (i = 0; i < n; i++)
			{
				Noeud no = racine;
				do
				{
					no.CalculMeilleurFils(a,b);
					no = no.MeilleurFils();

				} while (no.p.Eval == Resultat.indetermine);


				Resultat res = no.p.Eval;

				int re = p.j1aletrait ? 0 : 1;

				if (res == Resultat.j1gagne) { re = 1; }
				if (res == Resultat.j0gagne) { re = 0; }

				while (no != null)
				{
					no.cross += 1;
					no.win += re;
					no = no.pere;
				}
				if (sw.ElapsedMilliseconds > temps) { break; }
			}
			if (nbcoups == 0) { nbiter += i; nbcoups++; }

			racine.CalculMeilleurFils(a,b);
			if (affichage)
			{
				Console.WriteLine("{0} itérations en {1} ms, {2} nœuds visités", i, sw.ElapsedMilliseconds, Noeud.compteur);
				Console.WriteLine(racine);
			}

			int rep=racine.indiceMeilleurFils;
			racine = racine.MeilleurFils();
			racine.pere = null;
			return  rep;
		}
	}

	public class JMCTSoff : Joueur
	{
		public static Random gen = new Random();
		static Stopwatch sw = new Stopwatch();

		float a,b;
		int n;
		int temps;

		int nbiter, nbcoups;

		public override float Nbitmoy
		{
			get
			{
				return (1F * nbiter) / nbcoups;
			}
		}

		Noeud racine;

		public JMCTSoff(float a,float b, int n, int temps = int.MaxValue)
		{
			this.a = a; this.b = b; this.n = n; this.temps = temps;
		}

		public override string ToString()
		{
			return string.Format("JMCTSoff[a={0}  b={1}  n={2}  temps={3}]", a, b, n,temps);
		}

		public override void NouvellePartie()
		{
			racine = null; nbiter = 0; nbcoups = 0;
		}

		Resultat JeuHasard(Position p)
		{

			Resultat res = p.Eval;
			while (res == Resultat.indetermine)
			{
				p = p.PositionFille(gen.Next(0, p.NbCoups));
				res = p.Eval;
			}
			return res;
		}

		public override int Jouer(Position p, bool affichage)
		{
			sw.Restart();
			if (racine == null) { racine = new Noeud(null, p); }
			else {
				int k = 0;
				for (k = 0; k < racine.fils.Length; k++)
				{
					if (p.Equals(racine.p.PositionFille(k))) break;
				}
				racine = racine.InstancierFils(k);
				racine.pere = null;
			}
			int i = 0;
			for (i = 0; i < n; i++)
			{
				Noeud no = racine;
				do
				{
					no.CalculMeilleurFils(a,b);
					no = no.MeilleurFils();

				} while (no.fils.Length > 0 && no.cross > 0);


				Resultat res = JeuHasard(no.p);

				int re = p.j1aletrait ? 0 : 1;

				if (res == Resultat.j1gagne) { re = 1; }
				if (res == Resultat.j0gagne) { re = 0; }

				while (no != null)
				{
					no.cross += 1;
					no.win += re;
					no = no.pere;
				}
				if (sw.ElapsedMilliseconds > temps) { no=null;break; };
			}

			if (nbcoups == 0) { nbiter += i; nbcoups++;}

			racine.CalculMeilleurFils(a,b);
			if (affichage)
			{
				Console.WriteLine("{0} itérations en {1} ms", i, sw.ElapsedMilliseconds);
				Console.WriteLine(racine);
			}
			int rep = racine.indiceMeilleurFils;
			racine = racine.MeilleurFils();
			racine.pere = null;
			return rep;
		}
	}
}
